"""Exceptions"""

__copyright__ = "Copyright (C) 2014 Ivan D Vasin"
__docformat__ = "restructuredtext"

import exceptions as _py_exc


class Exception(_py_exc.Exception):
    pass


class Error(RuntimeError, Exception):
    pass


class UnsupportedUniversalSetOperation(Error):

    """
    A finite set operation was attempted on a universal set

    :param operation:
        The attempted operation.
    :type operation: :obj:`str`

    :param message:
        A message that describes the error.
    :type message: :obj:`str` or null

    """

    def __init__(self, operation, set, message=None, *args):
        super(UnsupportedUniversalSetOperation, self)\
         .__init__(operation, set, message, *args)
        self._message = message
        self._operation = operation
        self._set = set

    def __str__(self):
        message = '{} is unsupported by the universal set {!r}'\
                   .format(self.operation, self.set)
        if self.message:
            message += ': ' + self.message
        return message

    @property
    def message(self):
        return self._message

    @property
    def operation(self):
        return self._operation

    @property
    def set(self):
        return self._set



"""Universalizable set data types"""

__copyright__ = "Copyright (C) 2014 Ivan D Vasin"
__docformat__ = "restructuredtext"

from abc import ABCMeta as _ABCMeta, abstractmethod as _abstractmethod
from collections import MutableSet as _MutableSet, Set as _Set
from functools import total_ordering as _total_ordering
from itertools import product as _product

from . import _misc
from .. import _exc


@_total_ordering
class uset_abc(object):

    __metaclass__ = _ABCMeta

    def __init__(self, items=None):
        if not items:
            self._items = self._items_class()(())
        elif isinstance(self, self._frozen_class()) \
             and isinstance(items, self._frozen_class()) \
             and isinstance(items._items, self._items_class()):
            self._items = items._items
        elif (isinstance(items, (self._frozen_class(),
                                 self._unfrozen_class()))
              and not items.isfinite) \
             or (isinstance(items, basestring) and items == '*'):
            self._items = self._universe()
        else:
            self._items = self._items_class()(items)

    def __and__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not self.isfinite:
            return other
        elif not other.isfinite:
            return self
        else:
            return self.__class__(self._items & other._items)

    def __contains__(self, item):
        return not self.isfinite or item in self._items

    def __eq__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        return self._items == other._items

    def __gt__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not other.isfinite:
            return False
        elif not self.isfinite:
            return True
        else:
            return self._items < other._items

    def __iter__(self):
        if not self.isfinite and self._universe() == self._DEFAULT_UNIVERSE:
            raise _exc.UnsupportedUniversalSetOperation('iteration', set=self)
        else:
            return iter(self._items)

    def __len__(self):
        if not self.isfinite:
            return float('inf')
        else:
            return len(self._items)

    def __lt__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not self.isfinite:
            return False
        elif not other.isfinite:
            return True
        else:
            return self._items < other._items

    def __nonzero__(self):
        return bool(self._items)

    def __or__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not self.isfinite:
            return self
        elif not other.isfinite:
            return other
        else:
            return self.__class__(self._items | other._items)

    def __repr__(self):
        return '{}({!r})'\
                .format(self.__class__.__name__,
                        self._items if self.isfinite else '*')

    def __str__(self):
        return '{{{}}}'.format(', '.join(sorted(str(item)
                                                for item in self._items))) \
                   if self.isfinite \
                   else '*'

    def __sub__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not self.isfinite:
            raise _exc.UnsupportedUniversalSetOperation('set difference',
                                                        set=self)
        elif not other.isfinite:
            return self.__class__()
        else:
            return self.__class__(self._items - other._items)

    def __xor__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not (self.isfinite or other.isfinite):
            return self.__class__()
        elif not self.isfinite:
            raise _exc.UnsupportedUniversalSetOperation('XOR', set=self)
        elif not other.isfinite:
            raise _exc.UnsupportedUniversalSetOperation('XOR', set=other)
        else:
            return self.__class__(self._items ^ other._items)

    def copy(self):
        return self.__class__(self._items.copy())

    @_abstractmethod
    def frozen(self):
        pass

    def isdisjoint(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not self.isfinite:
            return not other._items
        elif not other.isfinite:
            return not self._items
        else:
            return self._items.isdisjoint(other._items)

    @property
    def isfinite(self):
        """

        .. note:: **TODO:**
            refactor as :attr:`!isuniversal` with opposite semantics

        """
        return self._items != self._universe()

    def set(self):
        if not self.isfinite:
            raise _exc.UnsupportedUniversalSetOperation\
                   ('conversion to a finite set', set=self)
        else:
            return self._items

    @_abstractmethod
    def unfrozen(self):
        pass

    @_abstractmethod
    def unfrozen_copy(self):
        pass

    _DEFAULT_UNIVERSE = object()

    @classmethod
    def _frozen_class(cls):
        return frozenuset

    @classmethod
    @_abstractmethod
    def _items_class(cls):
        pass

    @classmethod
    def _unfrozen_class(cls):
        return uset

    @classmethod
    def _universe(cls):
        return cls._DEFAULT_UNIVERSE


class frozenuset(uset_abc, _Set):

    """An immutable universalizable set

    .. seealso:: :class:`uset`

    """

    def __hash__(self):
        return hash(self._items)

    def __str__(self):
        return '=' + super(frozenuset, self).__str__()

    def frozen(self):
        return self

    @classmethod
    def _items_class(cls):
        return _misc.frozenset

    def unfrozen(self):
        return self._unfrozen_class()(self)

    def unfrozen_copy(self):
        return self.unfrozen()


class uset(uset_abc, _MutableSet):

    """A universalizable set

    A universalizable set can be universal---that is, it can represent the
    set of all possible items that it supports.

    .. note:: **TODO:**
        support set difference and XOR by storing and operating on
        compositions

    :param items:
        The items.  The value ``'*'`` has special meaning, denoting all
        possible items.
    :type items:
        :class:`uset` or :class:`frozenuset` or ~{:obj:`object`}
        or :obj:`str` or null

    """

    def __iand__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not self.isfinite:
            self._items = other._items
        elif not other.isfinite:
            pass
        else:
            self._items &= other._items

        return self

    def __ior__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not self.isfinite:
            pass
        elif not other.isfinite:
            self._items = other._items
        else:
            self._items |= other._items

        return self

    def __isub__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not self.isfinite:
            raise _exc.UnsupportedUniversalSetOperation\
                   ('in-place set difference', set=self)
        elif not other.isfinite:
            self.clear()
        else:
            self._items -= other._items

        return self

    def __ixor__(self, other):

        if not isinstance(other, self.__class__):
            other = self.__class__(other)

        if not (self.isfinite or other.isfinite):
            self.clear()
        elif not self.isfinite:
            raise _exc.UnsupportedUniversalSetOperation\
                   ('in-place XOR with a finite set', set=self)
        elif not other.isfinite:
            raise _exc.UnsupportedUniversalSetOperation\
                   ('in-place XOR with a finite set', set=other)
        else:
            self._items -= other._items

        return self

    def add(self, item):
        if not self.isfinite:
            pass
        else:
            self._items.add(item)

    def clear(self):
        self._items = self._items_class()()

    def discard(self, item):
        if not self.isfinite:
            raise _exc.UnsupportedUniversalSetOperation('discarding an item',
                                                        set=self)
        else:
            self._items.discard(item)

    def frozen(self):
        return self._frozen_class()(self)

    def unfrozen(self):
        return self

    def unfrozen_copy(self):
        return self.copy()

    @classmethod
    def _items_class(cls):
        return set


class usetset_abc(uset_abc):

    __metaclass__ = _ABCMeta

    def __init__(self, items=()):
        if not items or isinstance(items, (self._frozen_class(),
                                           self._unfrozen_class())) \
                     or (isinstance(items, basestring) and items == '*'):
            super(usetset_abc, self).__init__(items)
        else:
            self._items = self._items_class()(self._normalized_item(item)
                                              for item in items)

    def all_contain(self, item):
        return self.isfinite \
               and self._items \
               and all(item in own_item for own_item in self._items)

    def all_lt(self, item):
        return self.isfinite \
               and self._items \
               and all(own_item < item for own_item in self._items)

    def all_lte(self, item):
        return self.isfinite \
               and self._items \
               and all(own_item <= item for own_item in self._items)

    def all_gt(self, item):
        return self.isfinite \
               and self._items \
               and all(own_item > item for own_item in self._items)

    def all_gte(self, item):
        return self.isfinite \
               and self._items \
               and all(own_item >= item for own_item in self._items)

    def any_contain(self, item):
        return not self.isfinite or any(item in own_item
                                        for own_item in self._items)

    def any_eq(self, item):
        return not self.isfinite or any(own_item == item
                                        for own_item in self._items)

    def any_lt(self, item):
        return not self.isfinite or any(own_item < item
                                        for own_item in self._items)

    def any_lte(self, item):
        return not self.isfinite or any(own_item <= item
                                        for own_item in self._items)

    def any_gt(self, item):
        return not self.isfinite or any(own_item > item
                                        for own_item in self._items)

    def any_gte(self, item):
        return not self.isfinite or any(own_item >= item
                                        for own_item in self._items)

    def intersection_product(self, other):

        if not isinstance(other, usetset_abc):
            other = self.__class__(other)

        if self and other:
            return self.__class__({self._item_class()(item & other_item)
                                   for item, other_item
                                   in _product(self._items, other._items)})
        else:
            return self.__class__()

    def union_product(self, other):

        if not isinstance(other, usetset_abc):
            other = self.__class__(other)

        if self and other:
            if not self.isfinite or not other.isfinite:
                return self.__class__('*')
            else:
                return self.__class__({self._item_class()(item | other_item)
                                       for item, other_item
                                       in _product(self._items, other._items)})
        elif self:
            return self
        else:
            return other

    @classmethod
    def _frozen_class(cls):
        return frozenusetset

    @classmethod
    @_abstractmethod
    def _item_class(cls):
        pass

    @classmethod
    def _normalized_item(cls, item):
        item_class = cls._item_class()
        return item if isinstance(item, item_class) else item_class(item)

    @classmethod
    def _unfrozen_class(cls):
        return usetset


class frozenusetset(usetset_abc, frozenuset):

    """An immutable universalizable set of immutable sets

    .. seealso:: :class:`frozenuset`

    """

    @classmethod
    def _item_class(cls):
        return _misc.frozenset


class usetset(usetset_abc, uset):

    """A universalizable set of immutable sets

    .. seealso:: :class:`uset`

    """

    def add(self, item):
        if not self.isfinite:
            pass
        else:
            self._items.add(self._normalized_item(item))

    @classmethod
    def _item_class(cls):
        return _misc.frozenset



# -*- coding: utf-8 -*-

"""API documentation build configuration file.

This file is :func:`execfile`\\ d with the current directory set to its
containing dir.

Note that not all possible configuration values are present in this
autogenerated file.

All configuration values have a default; values that are commented out
serve to show the default.

"""

from datetime import date as _date
import re as _re
from textwrap import dedent as _dedent

import docutils.parsers.rst as _rst


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.coverage',
              'sphinx.ext.doctest',
              'sphinx.ext.ifconfig',
              'sphinx.ext.intersphinx',
              'sphinx.ext.mathjax',
              'sphinx.ext.todo',
              'sphinx.ext.viewcode',
              'sphinxcontrib.cheeseshop',
              ]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u''
author = u''
copyright = u'{} {}'.format(_date.today().year, author)
description = u''

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ''
# The full version, including alpha/beta/rc tags.
release = ''

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
#add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

rst_prolog = \
    '''\
    .. role:: bash(code)
        :language: bash

    .. role:: python(code)
        :language: python
    '''
rst_prolog = _dedent(rst_prolog)

nitpicky = True

# FIXME: encapsulate this in a Sphinx extension.  make ``rfc_uri_tmpl`` a
#     Sphinx config setting
rfc_uri_tmpl = 'https://tools.ietf.org/html/rfc{}.html'
def rfc_role(role, rawtext, text, lineno, inliner, options={}, content=[]):

    _rst.roles.set_classes(options)

    rfcrefpattern = r'(?:(?P<displaytext>[^<]*)'\
                     r' <)?(?P<refname>[^>]*)(?(displaytext)>|)'
    match = _re.match(rfcrefpattern, _rst.roles.utils.unescape(text))
    if match:
        rfcnum, anchorsep, anchor = match.group('refname').partition('#')
        try:
            rfcnum = int(rfcnum)
            if rfcnum <= 0:
                raise ValueError
        except ValueError:
            message = \
                inliner\
                 .reporter\
                 .error('invalid RFC number {!r}; expected a positive integer'
                         .format(rfcnum),
                        line=lineno)
            problem = inliner.problematic(rawtext, rawtext, message)
            return [problem], [message]

        uri = rfc_uri_tmpl.format(rfcnum)
        if anchor:
            uri += anchorsep + anchor

        displaytext = match.group('displaytext')
        if displaytext:
            refnode = _rst.nodes.reference(rawtext, displaytext, refuri=uri,
                                           **options)

        else:
            displaytext = 'RFC {}'.format(rfcnum)
            if anchor:
                displaytext += ' ' + anchor.replace('-', ' ')

            strongnode = _rst.nodes.strong(rawtext, displaytext)
            refnode = _rst.nodes.reference('', '', strongnode, refuri=uri,
                                           **options)

        return [refnode], []

    else:
        message = \
            inliner\
             .reporter\
             .error('invalid RFC reference {!r}'.format(text), line=lineno)
        problem = inliner.problematic(rawtext, rawtext, message)
        return [problem], [message]
_rst.roles.register_local_role('rfc', rfc_role)


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'nature'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = '{}-doc'.format(project)


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
  ('index', '{}.tex'.format(project), u'{} documentation'.format(project),
   author, 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [('index', project.lower(), u'{} documentation'.format(project),
              [author], 1)]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [('index', project, u'{} documentation'.format(project),
                      author, project, description, 'Miscellaneous')]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'


# -- Options for Epub output ---------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The language of the text. It defaults to the language option
# or en if the language is not set.
#epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
#epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#epub_identifier = ''

# A unique identification for the text.
#epub_uid = ''

# A tuple containing the cover image and cover page html template filenames.
#epub_cover = ()

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
#epub_pre_files = []

# HTML files shat should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
#epub_post_files = []

# A list of files that should not be packed into the epub file.
#epub_exclude_files = []

# The depth of the table of contents in toc.ncx.
#epub_tocdepth = 3

# Allow duplicate toc entries.
#epub_tocdup = True


# -- Options for intersphinx ---------------------------------------------------

intersphinx_mapping = \
    {'python': ('http://docs.python.org/2/', None),
     'sphinx': ('http://sphinx.readthedocs.org/en/latest/', None),
     }


# -- Options for autodoc -------------------------------------------------------

autodoc_default_flags = ['show-inheritance']



"""Mapping data types

A mapping is a set of *(key, value)* pairs.  A mapping data type is a
container that associates zero or more objects with each item of some
collection.

Despite the fact that a mapping is fundamentally a set,
:class:`collections.Mapping` does not support the API of
:class:`collections.Set`.

"""

__copyright__ = "Copyright (C) 2014 Ivan D Vasin"
__docformat__ = "restructuredtext"

from ._misc import *



"""Set data types

A set is a category of which every object is either a member or not.  A
set data type is a container whose items are distinct and unordered.

"""

__copyright__ = "Copyright (C) 2014 Ivan D Vasin"
__docformat__ = "restructuredtext"

from ._misc import *
from ._universalizable import *



"""Container data types

This is an extension of :mod:`collections` and the built-in mapping_,
sequence_, and set_ data types.


.. _mapping: \
    http://docs.python.org/2/library/stdtypes.html#mapping-types-dict

.. _sequence: \
    http://docs.python.org/2/library/stdtypes.html#sequence-types-str-unicode-list-tuple-bytearray-buffer-xrange

.. _set: \
    http://docs.python.org/2/library/stdtypes.html#set-types-set-frozenset

"""

__version__ = "0.2.3"
__copyright__ = "Copyright (C) 2014 Ivan D Vasin"
__credits__ = ["Ivan D Vasin"]
__maintainer__ = "Ivan D Vasin"
__email__ = "nisavid@gmail.com"
__docformat__ = "restructuredtext"

from ._exc import *
from ._mappings import *
from ._sequences import *
from ._sets import *



"""Miscellaneous mapping data types"""

__copyright__ = "Copyright (C) 2014 Ivan D Vasin"
__docformat__ = "restructuredtext"

from collections \
    import Mapping as _Mapping, MutableMapping as _MutableMapping, \
           OrderedDict as _OrderedDict
from functools import reduce as _reduce
from operator import xor as _xor

import multimap as _multimap
from spruce.types import require_isinstance as _require_isinstance


class _mdict_base(_multimap.MultiMap):

    def __init__(self, mapping_or_items=None):
        if mapping_or_items is None:
            super(_mdict_base, self).__init__(())
        elif hasattr(mapping_or_items, 'allitems'):
            super(_mdict_base, self).__init__(mapping_or_items.allitems())
        else:
            super(_mdict_base, self).__init__(mapping_or_items)

    def __repr__(self):
        # FIXME: make printed representations behave similarly to those of
        #   :obj:`list` and :obj:`dict`, converting the delimiters of long
        #   containers from ', ' to ',\n'
        return '{}({!r})'.format(self.__class__.__name__, self.allitems())

    def __str__(self):
        # FIXME: make printed representations behave similarly to those of
        #   :obj:`list` and :obj:`dict`, converting the delimiters of long
        #   containers from ', ' to ',\n'
        return '+{{{}}}'.format(', '.join('{}: {}'.format(key, value)
                                          for key, value in self.allitems()))


class _odict_base(_OrderedDict):

    def __repr__(self):
        # FIXME: make printed representations behave similarly to those of
        #   :obj:`list` and :obj:`dict`, converting the delimiters of long
        #   containers from ', ' to ',\n'
        return '{}({!r})'.format(self.__class__.__name__, self.items())

    def __str__(self):
        # FIXME: make printed representations behave similarly to those of
        #   :obj:`list` and :obj:`dict`, converting the delimiters of long
        #   containers from ', ' to ',\n'
        return '>{{{}}}'.format(', '.join('{}: {}'.format(key, value)
                                          for key, value in self.items()))


class _omdict_base(_mdict_base):
    # NOTE: this is nearly identical to :class:`_mdict_base` because that
    #   class's underlying implementation happens to be ordered
    def __str__(self):
        return '>' + super(_omdict_base, self).__str__()


class _typeddict_base(_Mapping):

    def __init__(self, mapping_or_items=None, keytype=None, key_converter=None,
                 valuetype=None, value_converter=None):

        if not keytype:
            raise TypeError('missing key type')

        if not valuetype:
            raise TypeError('missing value type')

        self._key_converter = key_converter if key_converter is not None \
                                            else keytype
        self._keytype = keytype
        self._value_converter = value_converter \
                                    if value_converter is not None \
                                    else valuetype
        self._valuetype = valuetype

        if mapping_or_items:
            if isinstance(mapping_or_items, typeddict) \
                   and issubclass(mapping_or_items.keytype, self.keytype) \
                   and issubclass(mapping_or_items.valuetype, self.valuetype):
                self._dict = self._dicttype()(mapping_or_items._dict)
            else:
                self._dict = self._dicttype()()
                if isinstance(mapping_or_items, _Mapping):
                    items = mapping_or_items.items()
                else:
                    items = mapping_or_items
                for key, value in items:
                    self._setitem(key, value)
        else:
            self._dict = self._dicttype()()

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        # FIXME: make printed representations behave similarly to those of
        #   :obj:`list` and :obj:`dict`, converting the delimiters of long
        #   containers from ', ' to ',\n'
        return '{}({!r}, keytype={!r}, valuetype={!r})'\
                .format(self.__class__.__name__, self._dict, self.keytype,
                        self.valuetype)

    def __str__(self):
        # FIXME: make printed representations behave similarly to those of
        #   :obj:`list` and :obj:`dict`, converting the delimiters of long
        #   containers from ', ' to ',\n'
        return '{{{}}}'.format(', '.join('{}: {}'.format(key, value)
                                         for key, value in self.items()))

    @property
    def key_converter(self):
        return self._key_converter

    @property
    def keytype(self):
        return self._keytype

    @property
    def value_converter(self):
        return self._value_converter

    @property
    def valuetype(self):
        return self._valuetype

    @classmethod
    def _dicttype(cls):
        return dict

    def _normalized_key(self, key):
        try:
            _require_isinstance(key, self.keytype)
        except ValueError:
            if self.key_converter:
                return self.key_converter(key)
            else:
                raise
        else:
            return key

    def _normalized_value(self, value):
        try:
            _require_isinstance(value, self.valuetype)
        except ValueError:
            if self.value_converter:
                return self.value_converter(value)
            else:
                raise
        else:
            return value

    def _setitem(self, key, value):
        self._dict[self._normalized_key(key)] = self._normalized_value(value)


class defaultmapping(_MutableMapping):

    def __init__(self, mapping, default_factory):
        self._default_factory = default_factory
        self._mapping = mapping

    def __delitem__(self, key):
        del self.mapping[key]

    def __getitem__(self, key):
        try:
            return self.mapping[key]
        except KeyError:
            value = self.default_factory()
            self.mapping[key] = value
            return value

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

    def __repr__(self):
        return '{}({!r}, {!r})'.format(self.__class__.__name__, self.mapping,
                                       self.default_factory)

    def __setitem__(self, key, value):
        self.mapping[key] = value

    def __str__(self):
        return '{}({}, {})'.format(self.__class__.__name__, self.mapping,
                                   self.default_factory)

    @property
    def default_factory(self):
        return self._default_factory

    @property
    def mapping(self):
        return self._mapping


class frozendict(_Mapping):

    """An immutable mapping"""

    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)
        self._hash = None

    def __getitem__(self, key):
        return self._dict[key]

    def __hash__(self):
        if self._hash is None:
            self._hash = _reduce(_xor,
                                 (hash(pair) for pair in self.iteritems()))
        return self._hash

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.items())

    def __str__(self):
        return str(self._dict)


class frozenmdict(_mdict_base):

    """An immutable multimap

    .. seealso:: :class:`multimap.MultiMap` from :pypi:`multimap`

    :param mapping_or_items:
        A mapping or a sequence of *(key, value)* items.
    :type mapping_or_items: ~+{:obj:`object`: :obj:`object`} or null

    """

    def __init__(self, mapping_or_items=None):
        super(frozenmdict, self).__init__(mapping_or_items)
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = _reduce(_xor,
                                 (hash(pair) for pair in self.iterallitems()))
        return self._hash

    def __str__(self):
        return '=' + super(frozenmdict, self).__str__()


class frozenodict(_Mapping):

    """An immutable ordered mapping

    .. seealso:: :class:`odict`

    """

    def __init__(self, *args, **kwargs):
        self._hash = None
        self._odict = odict(*args, **kwargs)

    def __getitem__(self, key):
        return self._odict[key]

    def __hash__(self):
        if self._hash is None:
            self._hash = _reduce(_xor,
                                 (hash(pair) for pair in self.iteritems()))
        return self._hash

    def __iter__(self):
        return iter(self._odict)

    def __len__(self):
        return len(self._odict)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.items())

    def __str__(self):
        return '=' + str(self._odict)


class frozenomdict(_omdict_base):

    """An immutable ordered multimap

    .. seealso:: :class:`multimap.MultiMap` from :pypi:`multimap`

    :param mapping_or_items:
        A mapping or a sequence of *(key, value)* items.
    :type mapping_or_items: ~+{:obj:`object`: :obj:`object`} or null

    """

    def __init__(self, mapping_or_items=None):
        super(frozenomdict, self).__init__(mapping_or_items)
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = _reduce(_xor,
                                 (hash(pair) for pair in self.iterallitems()))
        return self._hash

    def __str__(self):
        return '=' + super(frozenomdict, self).__str__()


class frozentypeddict(_typeddict_base):

    """An immutable typed mapping

    .. seealso:: :class:`typeddict`

    """

    def __init__(self, mapping_or_items=None, keytype=None, key_converter=None,
                 valuetype=None, value_converter=None):
        super(frozentypeddict, self)\
         .__init__(mapping_or_items, keytype=keytype,
                   key_converter=key_converter, valuetype=valuetype,
                   value_converter=value_converter)
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            self._hash = _reduce(_xor,
                                 (hash(pair) for pair in self.iteritems()),
                                 hash(self._dicttype()))
        return self._hash

    def __str__(self):
        return '=' + super(frozentypeddict, self).__str__()


class frozentypedmdict(frozentypeddict):
    """An immutable typed multimap

    .. seealso:: :class:`frozentypeddict`, :class:`frozenmdict`

    .. note:: **TODO:**
        support the full API of :class:`frozenmdict`

    """
    @classmethod
    def _dicttype(cls):
        return frozenmdict


class frozentypedodict(frozentypeddict):
    """An immutable typed ordered mapping

    .. seealso:: :class:`frozentypeddict`, :class:`frozenodict`

    """
    @classmethod
    def _dicttype(cls):
        return frozenodict


class frozentypedomdict(frozentypeddict):
    """An immutable typed ordered multimap

    .. seealso:: :class:`frozentypeddict`, :class:`frozenomdict`

    .. note:: **TODO:**
        support the full API of :class:`frozenomdict`

    """
    @classmethod
    def _dicttype(cls):
        return frozenomdict


class mdict(_mdict_base, _multimap.MutableMultiMap):
    """A multimap

    If a *mapping_or_items* is given, then the multimap is initialized with
    the items in it.  If the *mapping_or_items* is a multimap (that is, if
    it has a method :meth:`!allitems` that, when called, returns a sequence
    of all *(key, value)* items), then the new multimap is initialized with
    all of the items in *mapping_or_items*.  This is different from the
    behavior of :class:`multimap.MultiMap`, which treats all mappings the
    same.

    .. seealso:: :class:`multimap.MutableMultiMap` from :pypi:`multimap`

    :param mapping_or_items:
        A mapping or a sequence of *(key, value)* items.
    :type mapping_or_items: ~+{:obj:`object`: :obj:`object`} or null

    """
    pass


class odict(_odict_base):
    """An ordered mapping

    .. seealso:: :class:`collections.OrderedDict`

    """
    pass


class omdict(_omdict_base, _multimap.MutableMultiMap):
    """An ordered multimap

    .. seealso:: :class:`mdict`

    """
    pass


class typeddict(_typeddict_base, _MutableMapping):

    """A typed mapping

    This is a mapping from keys of a particular type to values of a
    particular type.  In addition to the basic mapping functionality, it
    validates its keys and values.

    The keys and values assigned to this mapping are validated using
    :func:`spruce.lang.require_isinstance()
    <spruce.lang._datatypes._checking.require_isinstance>`.  It is
    recommended to specify *keytype* and *valuetype* as subclasses of
    :class:`spruce.lang.AnnotatedType
    <spruce.lang._datatypes._annotated.AnnotatedType>` to ensure good error
    messages in case validation fails and normalization is not available.

    If validation fails for a key or value, then it is optionally normalized
    before being added.  This happens for keys as follows (values are
    handled similarly using the corresponding arguments).  If a key *K* is
    assigned for which :samp:`isinstance({K}, {keytype})`
    returns false, then the following operation takes place:

      * If *key_converter* is neither null nor false, then the assigned key
        is :samp:`{key_converter}({K})`.

      * If *key_converter* is null, then the assigned key is
        :samp:`{keytype}({K})`.

      * If *key_converter* is not null but false, then an error is raised as
        by :samp:`spruce.lang.require_isinstance({K}, {keytype})`.

    :param mapping_or_items:
        A mapping or item sequence of initial items.
    :type mapping_or_items:
        ~{(~\ *keytype* or ~\ *key_converter*): \
              (~\ *valuetype* or ~\ *value_converter*)}
        or null

    :param type keytype:
        The key type.

    :param type valuetype:
        The value type.

    :param key_converter:
        A callable that normalizes invalid keys.
    :type key_converter: :obj:`object` -> :obj:`object`

    :param value_converter:
        A callable that normalizes invalid values.
    :type value_converter: :obj:`object` -> :obj:`object`

    :raise TypeError:
        Raised if:

          * one or more of the keys of *mapping_or_items* is not an
            instance of *keytype* and *key_converter* is false but not
            null; or

          * one or more of the values of *mapping_or_items* is not an
            instance of *valuetype* and *value_converter* is false but not
            null.

    :raise:
        Other exceptions may be raised by *key_converter*, *keytype*,
        *value_converter*, or *valuetype* if they are called by the
        normalization behavior described above.

    """

    def __delitem__(self, key):
        del self._dict[key]

    def __setitem__(self, key, value):
        self._setitem(key, value)


class typedmdict(typeddict):
    """A typed multimap

    .. seealso:: :class:`typeddict`, :class:`mdict`

    .. note:: **TODO:**
        support the full API of :class:`mdict`

    """
    @classmethod
    def _dicttype(cls):
        return mdict


class typedodict(typeddict):
    """A typed ordered mapping

    .. seealso:: :class:`typeddict`, :class:`odict`

    """
    @classmethod
    def _dicttype(cls):
        return odict


class typedomdict(typeddict):
    """A typed ordered multimap

    .. seealso:: :class:`typeddict`, :class:`omdict`

    .. note:: **TODO:**
        support the full API of :class:`omdict`

    """
    @classmethod
    def _dicttype(cls):
        return omdict



"""Miscellaneous set data types"""

__copyright__ = "Copyright (C) 2014 Ivan D Vasin"
__docformat__ = "restructuredtext"

import __builtin__
from collections import Sequence as _Sequence, Set as _Set
from functools import reduce as _reduce
from operator import xor as _xor
import oset as _oset


class set(__builtin__.set):

    """A set

    .. seealso:: :class:`set <python:set>`

    """

    def __repr__(self):
        return '{{{}}}'.format(', '.join(repr(item) for item in self)) \
               if self else 'set()'

    def __str__(self):
        return '{{{}}}'.format(', '.join(str(item) for item in self)) \
               if self else 'set()'


class frozenoset(_Set, _Sequence):

    """An immutable ordered set

    .. seealso:: :mod:`oset` from :pypi:`oset`

    """

    def __init__(self, *args, **kwargs):
        self._hash = None
        self._oset = oset(*args, **kwargs)

    def __getitem__(self, key):
        return self._oset[key]

    def __hash__(self):
        if self._hash is None:
            self._hash = _reduce(_xor, (hash(item) for item in self))
        return self._hash

    def __len__(self):
        return len(self._oset)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, list(self))

    def __str__(self):
        return '=>{{{}}}'.format(', '.join(str(item) for item in self))


class frozenset(__builtin__.frozenset):

    """An immutable set.

    .. seealso:: :class:`frozenset <python:frozenset>`

    """

    def __repr__(self):
        return '={{{}}}'.format(', '.join(repr(item) for item in self))

    def __str__(self):
        return '={{{}}}'.format(', '.join(str(item) for item in self))


class oset(_oset.oset):

    """An ordered set.

    .. seealso:: :mod:`oset` from :pypi:`oset`

    """

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, list(self))

    def __str__(self):
        return '>{{{}}}'.format(', '.join(str(item) for item in self))



"""Sequence data types

A sequence is an ordered multiset.  A sequence data type is a container
in which the order of the items is significant and each item may appear
more than once.

Fixed-length sequences are also known as tuples; they are represented by
:obj:`tuple` and its derivations.  Variable-length sequences are also
known as lists; they are represented by :class:`collections.Sequence`
and its derivations.

Although a sequence is a multiset, a multiset is not a set of its
contents, so a sequence is also not a set of its contents.  Therefore
the sequence APIs (:obj:`tuple` and :class:`collections.Sequence`) do
not extend the API of :class:`collections.Set`.

Also, although a sequence can be interpreted as a mapping from integers
(the positions of its contents) to its contents, it is not fundamentally
defined so.  Therefore the sequence APIs do not extend the API of
:class:`collections.Mapping`.

"""

__copyright__ = "Copyright (C) 2014 Ivan D Vasin"
__docformat__ = "restructuredtext"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = "Copyright (C) 2014 Ivan D Vasin"
__credits__ = ["Ivan D Vasin"]
__maintainer__ = "Ivan D Vasin"
__email__ = "nisavid@gmail.com"
__docformat__ = "restructuredtext"

import io as _io
import os as _os
import re as _re

from setuptools import find_packages as _find_packages, setup as _setup


# helpers ---------------------------------------------------------------------


MOD_DOC_RE = _re.compile(r'\A(?:(?:\s*(?:\#[^$]*)?)\n)*'
                          r'(?:(?P<dquot>""")|(?P<squot>\'\'\'))\n?'
                          r'(?P<doc>.*)'
                          r'(?(dquot)"""|\'\'\')',
                         _re.DOTALL | _re.MULTILINE)

MOD_VERSION_RE = _re.compile(r'^__version__ = [\'"](?P<version>[^\'"]*)[\'"]$',
                             _re.MULTILINE)


def file_abspath_from_parts(*relpath_parts):
    return _os.path.join(_os.path.abspath(_os.path.dirname(__file__)),
                         *relpath_parts)


def shortdoc_from_doc(doc):
    def iter_shortdoc_lines():
        for line in doc.split(_os.linesep):
            if not line.strip():
                return

            if line[-1] in '.?!':
                line += ' '

            yield line
    return ' '.join(iter_shortdoc_lines())


def read_doc(mod_filepath):

    mod_text = _io.open(mod_filepath).read()
    match = MOD_DOC_RE.search(mod_text)

    if not match:
        raise RuntimeError('cannot find module docstring in file {!r} with'
                            ' pattern {!r}'
                            .format(mod_filepath, MOD_DOC_RE.pattern))

    return match.group('doc')


def read_version(mod_filepath):

    mod_text = _io.open(mod_filepath).read()
    mod_version_match = MOD_VERSION_RE.search(mod_text)

    if not mod_version_match:
        raise RuntimeError('cannot find module version string in file {!r}'
                            ' with pattern {!r}'
                            .format(mod_filepath, MOD_VERSION_RE.pattern))

    return mod_version_match.group('version')


# basics ----------------------------------------------------------------------

NAME_NOPREFIX = 'collections'

NAME_NOPREFIX_SAFE = _re.sub(r'[^\d\w]', '_',
                             _re.sub(r'^[^\w]', '_', NAME_NOPREFIX))

NAME = 'Spruce-' + NAME_NOPREFIX

PARENT_NAMESPACE_PKG_NAME = 'spruce'

ROOT_PKG_NAME = '.'.join((PARENT_NAMESPACE_PKG_NAME, NAME_NOPREFIX_SAFE))

ROOT_PKG_ABSPATH = file_abspath_from_parts(*(ROOT_PKG_NAME.split('.')
                                              + ['__init__.py']))

VERSION = read_version(ROOT_PKG_ABSPATH)

SITE_URI = ''

DOWNLOAD_URI = 'https://github.com/nisavid/spruce-collections'

DESCRIPTION = shortdoc_from_doc(read_doc(ROOT_PKG_ABSPATH))

README_FILE = 'README.rst'
README = _io.open(README_FILE, 'r').read()

CHANGES_FILE = 'CHANGES.rst'
CHANGES = _io.open(CHANGES_FILE, 'r').read()

LICENSE_FILE = 'LICENSE'
LICENSE = _io.open(LICENSE_FILE, 'r').read()

LONG_DESCRIPTION = '\n\n'.join((README, CHANGES))

TROVE_CLASSIFIERS = \
    ('Development Status :: 4 - Beta',
     'Intended Audience :: Developers',
     'License :: OSI Approved :: GNU Lesser General Public License v3'
      ' (LGPLv3)',
     'Operating System :: POSIX',
     'Programming Language :: Python :: 2.7',
     'Topic :: Software Development :: Libraries :: Python Modules',
     )


# dependencies ----------------------------------------------------------------

SETUP_DEPS = ()

INSTALL_DEPS = ('multimap', 'oset', 'spruce-types')

EXTRAS_DEPS = {}

TESTS_DEPS = ()

DEPS_SEARCH_URIS = ()


# packages --------------------------------------------------------------------

NAMESPACE_PKGS_NAMES = (PARENT_NAMESPACE_PKG_NAME,)

SCRIPTS_PKG_NAME = '.'.join((ROOT_PKG_NAME, 'scripts'))

TESTS_PKG_NAME = '.'.join((ROOT_PKG_NAME, 'tests'))


# entry points ----------------------------------------------------------------

STD_SCRIPTS_PKG_COMMANDS = {}

COMMANDS = {cmd: '{}.{}:{}'.format(SCRIPTS_PKG_NAME,
                                   script if isinstance(script, basestring)
                                          else script[0],
                                   'main' if isinstance(script, basestring)
                                          else script[1])
            for cmd, script in STD_SCRIPTS_PKG_COMMANDS.items()}

ENTRY_POINTS = {'console_scripts': ['{} = {}'.format(name, funcpath)
                                    for name, funcpath in COMMANDS.items()]}


if __name__ == '__main__':
    _setup(name=NAME,
           version=VERSION,
           url=SITE_URI,
           download_url=DOWNLOAD_URI,
           description=DESCRIPTION,
           long_description=LONG_DESCRIPTION,
           author=', '.join(__credits__),
           maintainer=__maintainer__,
           maintainer_email=__email__,
           license=LICENSE,
           classifiers=TROVE_CLASSIFIERS,
           setup_requires=SETUP_DEPS,
           install_requires=INSTALL_DEPS,
           extras_require=EXTRAS_DEPS,
           tests_require=TESTS_DEPS,
           dependency_links=DEPS_SEARCH_URIS,
           namespace_packages=NAMESPACE_PKGS_NAMES,
           packages=_find_packages(),
           test_suite=TESTS_PKG_NAME,
           include_package_data=True,
           entry_points=ENTRY_POINTS)
